package de.mhus.lib;

import java.util.Calendar;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import de.mhus.lib.logging.MLog;

public abstract class IProperties implements Iterable<Map.Entry<String,Object>> {

	/**
	 * Overwrite this function to provide values in string format.
	 * 
	 * @param name
	 * @return null if the property not exists or the property value.
	 */
	public abstract Object getProperty(String name);
	
	@Deprecated
	public int getProperty(String name, int def) {
		return getInt(name,def);
	}

	@Deprecated
	public boolean getProperty(String name, boolean def) {
		return getBoolean(name, def);
	}
	
	@Deprecated
	public String getProperty(String name, String def) {
		Object out = getProperty(name);
		if (out == null) return def;
		return String.valueOf(out);
	}

	public String getString(String name, String def) {
		Object out = getProperty(name);
		if (out == null) return def;
		return String.valueOf(out);
	}
	
	public boolean getBoolean(String name, boolean def) {
		Object out = getProperty(name);
		if (out == null) return def;
		if (out instanceof Boolean) return ((Boolean)out).booleanValue();
		return MCast.toboolean(String.valueOf(out), def);
	}
	
	public int getInt(String name, int def) {
		Object out = getProperty(name);
		if (out == null) return def;
		if (out instanceof Number) return ((Number)out).intValue();
		return MCast.toint(String.valueOf(out),def);
	}
	
	public long getLong(String name, long def) {
		Object out = getProperty(name);
		if (out == null) return def;
		if (out instanceof Number) return ((Number)out).longValue();
		return MCast.tolong(String.valueOf(out), def);
	}
	
	public float getFloat(String name, float def) {
		Object out = getProperty(name);
		if (out == null) return def;
		if (out instanceof Number) return ((Number)out).floatValue();
		return MCast.tofloat(String.valueOf(out), def);
	}
	
	public double getDouble(String name, double def) {
		Object out = getProperty(name);
		if (out == null) return def;
		if (out instanceof Number) return ((Number)out).doubleValue();
		return MCast.todouble(String.valueOf(out),def);
	}
	
	public Calendar getCalendar(String name) {
		Object out = getProperty(name);
		if (out == null) return null;
		if (out instanceof Calendar) return (Calendar)out;
		return MCast.toCalendar(String.valueOf(out));
	}

	public MDate getDate(String name) {
		try {
			return new MDate(getProperty(name));
		} catch (Throwable t) {}
		return null;
	}
	
	public void setString(String name, String value) throws MException {
		setProperty(name, value);
	}
	
	public void setInt(String name, int value) throws MException {
		setProperty(name, value);
	}
	
	public void setLong(String name, long value) throws MException {
		setProperty(name, value);
	}

	public void setDouble(String name, double value) throws MException {
		setProperty(name, value);
	}

	public void setFloat(String name, float value) throws MException {
		setProperty(name, value);
	}
	
	public void setBoolean(String name, boolean value) throws MException {
		setProperty(name, value);
	}
	
	public void setCalendar(String name, Calendar value) throws MException {
		setProperty(name, value);
	}
	
	public void setDate(String name, MDate value) throws MException {
		setProperty(name, value);
	}
	
	/**
	 * Return true if the property exists.
	 * 
	 * @param name
	 * @return
	 */
	public abstract boolean isProperty(String name);

	/**
	 * Remove the property field in the list of properties.
	 * 
	 * @param key
	 */
	public abstract void removeProperty(String key) throws MException;
	
	/**
	 * Overwrite this function to allow changes in properties.
	 * 
	 * @param key
	 * @param value
	 */
	public abstract void setProperty(String key, Object value) throws MException;
	
	/**
	 * Overwrite this function and return true if the property set can be edited.
	 * 
	 * @return
	 */
	public abstract boolean isEditable();
	
	/**
	 * 
	 */
	public abstract Set<String> keys();
	
	public Iterator<Map.Entry<String, Object>> iterator() {
		return new IPIterator();
	}
	
	private class IPIterator implements Iterator<Map.Entry<String, Object>> {

		private Iterator<String> keys;
		private String currentkey;

		IPIterator() {
			keys = keys().iterator();
		}
		
		@Override
		public boolean hasNext() {
			return keys.hasNext();
		}

		@Override
		public Entry<String, Object> next() {
			currentkey = keys.next();
			return new IPEntry(currentkey);
		}

		@Override
		public void remove() {
			try {
				removeProperty(currentkey);
			} catch (MException e) {
				MLog.t(e);
			}
		}
	}
	
	private class IPEntry implements Map.Entry<String, Object> {

		private String key;

		public IPEntry(String next) {
			key = next;
		}

		@Override
		public String getKey() {
			return key;
		}

		@Override
		public Object getValue() {
			return getProperty(key);
		}

		@Override
		public Object setValue(Object value) {
			Object old = getProperty(key);
			try {
				setProperty(key, value);
			} catch (MException e) {
				MLog.t(e);
			}
			return old;
		}
		
	}
}
