package de.mhus.lib;

import java.io.Closeable;
import java.io.File;
import java.util.HashMap;
import java.util.Map;
import java.util.Observer;

import de.mhus.lib.cao.CaoFactory;
import de.mhus.lib.config.ConfigUtil;
import de.mhus.lib.config.HashConfig;
import de.mhus.lib.config.IConfig;
import de.mhus.lib.console.Console;
import de.mhus.lib.jmx.MRemoteManager;
import de.mhus.lib.lang.MObject;
import de.mhus.lib.logging.ConsoleFactory;
import de.mhus.lib.logging.Log;
import de.mhus.lib.logging.MLog;
import de.mhus.lib.persistence.MPersistenceManager;
import de.mhus.lib.resources.MResources;
import de.mhus.lib.sql.DbPoolBundle;

public class MSingleton extends MObject {

	private static MSingleton singleton = new MSingleton();
	
	protected IConfig config = null;
	protected MActivator activator = new MActivator();
	protected MEventHandler<Observer> configChangeObserver = new MEventHandler<Observer>();
	protected Log logFactory = new ConsoleFactory();
	protected MHousekeeper housekeeper = null;
	protected CaoFactory caoFactory = null;
	protected DbPoolBundle dbPool = null;
	protected Map<String, Object> objects = new HashMap<String, Object>();
	protected MRemoteManager remoteManager = null;
	protected MArgs args = null;
	protected String defaultConfigFile = null;
	protected Console console = null;
	protected MPersistenceManager persistence = null;
	protected MThreadManager threadManager;
	protected int uniqueId = 0;
	protected MResources resources;
	protected File applicationDirectory;

	private String applicationName;

 

	public MSingleton() {
		this(singleton);
	}
	
	protected MSingleton(MSingleton singleton) {
		if (singleton!=null) {
			activator = singleton.activator;
			configChangeObserver = singleton.configChangeObserver;
			config = singleton.config;
			logFactory = singleton.logFactory;
			housekeeper = singleton.housekeeper;
			threadManager = singleton.threadManager;
			caoFactory = singleton.caoFactory;
			dbPool = singleton.dbPool;
			objects = singleton.objects;
			remoteManager = singleton.remoteManager;
			args = singleton.args;
			uniqueId = singleton.uniqueId;
			defaultConfigFile = singleton.defaultConfigFile;
			resources = singleton.resources;
			console = singleton.console;
			persistence = singleton.persistence;
			applicationDirectory = singleton.applicationDirectory;
			applicationName = singleton.applicationName;
		}
	}
	
	protected void destroy() {
		for (Object obj : objects.values() )
			if (obj instanceof Closeable) {
				try {
					((Closeable)obj).close();
				} catch (Throwable t) {}
			}
		objects = null;
		if (remoteManager != null) remoteManager.close();
		remoteManager = null;
		if (housekeeper!=null) housekeeper.finalize();
		housekeeper = null;
		threadManager = null;
		activator = null;
		config = null;
		configChangeObserver = null;
		logFactory = null;
		caoFactory = null;
		if (dbPool != null) dbPool.close();
		dbPool = null;
		singleton = new MSingleton(null);
		defaultConfigFile = null;
		
		if (persistence != null) persistence.close();
		persistence = null;
	}
	
	protected static void setInstance(MSingleton inst) {
		if (inst != null) singleton = inst;
	}
	
	/**
	 * Return the central instance of the singleton.
	 * 
	 * @return the singleton.
	 */
	public static MSingleton instance() {
		return singleton;
	}
	
	/**
	 * Return an config for the given object (class name). If not found it use the default config. It also will check conditions. If the
	 * condition fails it will return null.
	 * 
	 * @param obj
	 * @param def
	 * @return
	 */
	public synchronized IConfig getConfig(Object obj, IConfig def) {
		IConfig out = null;
		if (obj == null) {
			// nothing
		} else
		if (obj instanceof Class) {
			out = getConfig().getConfig(((Class<?>)obj).getCanonicalName());
		} else {
			out = getConfig().getConfig(obj.getClass().getCanonicalName());
		}
		if (out == null) out = def;
		if (out == null) return null;

		if (!ConfigUtil.checkCondition(out)) return null;
		return out;
	}
	
	/**
	 * Return the central configuration.
	 * @return
	 */
	public synchronized IConfig getConfig() {
		if (config == null)
			config = new HashConfig();
		return config;
	}
	
	/**
	 * Return the central activator.
	 * 
	 * @return
	 */
	public MActivator getActivator() {
		return activator;
	}

	/**
	 * Register a listener to get events if the config is changed. The reference in the list is weak. This
	 * means the listener must be hold on another variable too.
	 * 
	 * @param listener
	 */
	public void registerConfigListener(Observer listener) {
		configChangeObserver.registerWeak(listener);
	}

	/**
	 * Remove a registered config listener.
	 * 
	 * @param listener
	 */
	public void unregisterConfigListener(Observer listener) {
		configChangeObserver.unregister(listener);
	}
	
	/**
	 * Fire this if the config is changed.
	 */
	protected void fireConfigChanged() {
		for (Object listener : configChangeObserver.getListenersArray()) {
			((Observer)listener).update(null, config);
		}
	}
	
	/**
	 * Return the central log factory.
	 * 
	 * @return
	 */
	public Log getLog() {
		return logFactory;
	}

	/**
	 * Set the central log factory.
	 * 
	 * @param log
	 */
	public void setLogFactory(Log log) {
		this.logFactory = log;
		MLog.reset();
	}
	
	/**
	 * Return the central housekeeper.
	 * 
	 * @return
	 */
	public synchronized MHousekeeper getHousekeeper() {
		if (housekeeper == null)
			housekeeper = new MHousekeeper();
		return housekeeper;
	}

	/**
	 * Return the central thread pool manager.
	 * 
	 * @return
	 */
	public synchronized MThreadManager getThreadManager() {
		if (threadManager == null)
			threadManager = new MThreadManager();
		return threadManager;
	}

	/**
	 * Return a central cao factory.
	 * 
	 * @return
	 */
	public synchronized CaoFactory getCaoFactory() {
		if (caoFactory == null)
			caoFactory = new CaoFactory();
		return caoFactory;
	}
	
	/**
	 * Return a central database pool.
	 * 
	 * @return
	 */
	public synchronized DbPoolBundle getDbPool() {
		if (dbPool == null)
			dbPool = new DbPoolBundle();
		return dbPool;
	}
	
	/**
	 * Remember user objects in the singleton.
	 * 
	 * @param key The unique name of the user object
	 * @param object The object to store or null if the object should be removed.
	 * @return The last stored object.
	 */
	public Object setObject(String key, Object object) {
		if (key == null) return null;
		if (object == null) {
			return objects.remove(key);
		}
		return objects.put(key, object);
	}
	
	/**
	 * Get a user object by unique key.
	 * 
	 * @param key
	 * @return
	 */
	public Object getObject(String key) {
		return objects.get(key);
	}
	
	public synchronized MRemoteManager getJmxManager() {
		if (remoteManager == null)
			remoteManager = new MRemoteManager();
		return remoteManager;
	}
	
	public boolean isJmx() {
		return remoteManager != null;
	}
	
	public void setArguments(String[] args) {
		if (this.args != null) return;
		this.args = new MArgs(args);
		log().t("set args",this.args);
	}
	
	public MArgs getArguments() {
		return args;
	}

	public long nextUniqueId() {
		return uniqueId++;
	}
	
	public void setDefaultConfigFile(String name) {
		defaultConfigFile = name;
	}

	public MResources getResources() {
		if (resources == null) resources = new MResources();
		return resources;
	}
	
	public Console getConsole() {
		if (console == null) 
			console = Console.create();
		return console;
	}
	
	public void setConsole(Console console) {
		log().t("set console",console);
		this.console = console;
	}

	/**
	 * Compare the secret with a secret from the config file. This should be used
	 * for simple but needed security checks.
	 * 
	 * @param secret
	 * @return
	 */
	public boolean hasSuperPrivileges(String secret) {
		if (secret == null) return false;
		return secret.equals(MPassword.decode(MSingleton.instance().getConfig().getString("secret", null)));
	}
	
	/** 
	 * Return true if the persistence is activated for this application
	 * 
	 * @return
	 */
	public boolean isPersistence() {
		return persistence != null && persistence.isActive();
	}
	
	/**
	 * Return the persistence manager or null if no persistence is activated.
	 * 
	 * @return
	 */
	public MPersistenceManager getPersistenceManager() {
		return persistence;
	}

	public File getApplicationDirectory() {
		if (applicationDirectory == null) {
			String appDir = getConfig().getExtracted("application_directory");
			if (appDir == null) {
				appDir = MSystem.getTmpDirectory();
				appDir = appDir + File.separatorChar + MFile.normalize(getApplicationName());
			}
			applicationDirectory = new File(appDir);
		}
		return applicationDirectory;
	}

	public String getApplicationName() {
		if (applicationName == null) {
			String appName = getConfig().getExtracted("application_name");
			
			if (appName == null) {
				for (StackTraceElement[] stack : Thread.getAllStackTraces().values()) {
				    StackTraceElement main = stack[stack.length - 1];
				    if (main.getMethodName().equals("main")) {
				    	appName = main.getClassName();
				    	break;
				    }
				}
			}
			
			if (appName == null) {
				appName = "unknown";
			}
			applicationName = appName;
		}
		return applicationName;
	}
	
}
