package de.mhus.lib.adb.model;

import java.util.Calendar;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.Map.Entry;

import javax.smartcardio.ATR;

import de.mhus.lib.MException;
import de.mhus.lib.MString;
import de.mhus.lib.adb.DbManager;
import de.mhus.lib.adb.DbSchema;
import de.mhus.lib.adb.annotations.DbIndex;
import de.mhus.lib.adb.annotations.DbPersistent;
import de.mhus.lib.adb.annotations.DbTable;
import de.mhus.lib.config.HashConfig;
import de.mhus.lib.config.IConfig;
import de.mhus.lib.config.MConfigFactory;
import de.mhus.lib.lang.MObject;
import de.mhus.lib.lang.Raw;
import de.mhus.lib.sql.DbConnection;
import de.mhus.lib.sql.DbPrepared;
import de.mhus.lib.sql.DbResult;
import de.mhus.lib.sql.Dialect;
import de.mhus.lib.util.Rfc1738;

public abstract class Table extends MObject {

	protected Class<?> clazz;
	protected String registryName;
	protected DbManager manager;
	protected String name;
	protected String tableNameOrg;
	protected DbSchema schema;
	protected String tableName;
	protected HashMap<String,Field> fIndex = new HashMap<String, Field>();
	protected LinkedList<Field> fList = new LinkedList<Field>();
	private HashMap<String, LinkedList<Field>> iIdx = new HashMap<String, LinkedList<Field>>();
	protected LinkedList<Field> pk = new LinkedList<Field>();
	private DbPrepared sqlPrimary;
	private DbPrepared sqlInsert;
	private DbPrepared sqlUpdate;
	private DbPrepared sqlRemove;
	private LinkedList<Feature> features = new LinkedList<Feature>();
	protected IConfig attributes;

	public void init(DbManager manager, Class<?> clazz, String registryName, String tableName) {
		this.manager = manager;
		this.schema = manager.getSchema();
		this.clazz = clazz;
		this.registryName = registryName;
		this.tableName = tableName;
	}

	public void initDatabase(DbConnection con) throws Exception {
		
		DbTable table = clazz.getAnnotation(DbTable.class);
		if (tableName != null) {
			name = tableName;
		} else
		if (table == null || MString.isEmptyTrim(table.tableName())) {
			name = clazz.getSimpleName();
		} else {
			name = table.tableName();
		}
		
		if (table != null &&!MString.isEmptyTrim(table.attributes())) {
			attributes = MConfigFactory.getInstance().toConfig(table.attributes());
		} else {
			attributes = new HashConfig();
		}
		
		tableNameOrg = schema.getTableName(name);
		tableName = manager.getPool().getDialect().normalizeTableName(tableNameOrg);
		
		log().t("new table",name,tableName);
		
		parseFields();
		
		// features
		if (table != null && !MString.isEmptyTrim(table.features())) {
			for (String featureName : MString.split(table.features(),",")) {
				Feature feature = manager.getSchema().createFeature(manager, this, featureName);
				if (feature != null) features.add(feature);
			}
		}
		
		createTable(con);
		postInit();
	}

	protected abstract void parseFields() throws Exception;


	protected void addToIndex(String list, Field field) {
		for (String nr : MString.split(list, ",")) {
			LinkedList<Field> list2 = iIdx.get(nr);
			if (list2 == null) {
				list2 = new LinkedList<Field>();
				iIdx.put(nr, list2);
			}
			list2.add(field);
		}
	}

	protected void addField(Field field) {
		field.table = this;
		fIndex.put(field.createName, field);
		fList.add(field);
		if (field.isPrimary && field.isPersistent()) pk.add(field);
	}

	public void fillNameMapping(HashMap<String, Object> nameMapping) throws Exception {
		nameMapping.put("db." + name, new Raw(tableName));
		for (Field f : fList) {
			f.fillNameMapping(nameMapping);
		}
	}

	public void prepareCreate(Object object) throws Exception {
		for (Field f : fList) {
			f.prepareCreate(object);
		}
	}

	public void createObject(DbConnection con, Object object) throws Exception {
		
		for (Feature f : features)
			f.createObject(con,object);
		
		HashMap<String, Object> attributes = new HashMap<String, Object>();
		for (Field f : fList) {
			attributes.put(f.name, f.getFromTarget(object));
		}
		
		sqlInsert.getStatement(con).execute(attributes);
		
	}

	public void saveObject(DbConnection con, Object object) throws Exception {
		
		for (Feature f : features)
			f.saveObject(con,object);
				
		HashMap<String, Object> attributes = new HashMap<String, Object>();
		for (Field f : fList) {
			attributes.put(f.name, f.getFromTarget(object));
		}
		
		int c = sqlUpdate.getStatement(con).executeUpdate(attributes);
		if ( c != 1)
			throw new MException("update failed, updated objects " + c);
		
	}
	
	protected void postInit() throws MException {
		
		Collections.sort(pk, new Comparator<Field>() {

			@Override
			public int compare(Field o1, Field o2) {
				return o1.name.compareTo(o2.name);
			}
		});
		
		String sql = "SELECT * FROM " + tableName + " WHERE ";
		int nr = 0;
		for (Field f : pk) {
			sql+= (nr > 0 ? " AND " : "" ) + f.name + "=$" + nr + "$";
			nr++;
		}
		// TODO dialect.appendSqlLimit(0,1)
		
		sqlPrimary = manager.getPool().createStatement(sql);
		
		// ------
		
		sql = "INSERT INTO " + tableName + " (";
		nr = 0;
		for (Field f : fList) {
			if (f.isPersistent()) {
				if (nr > 0) sql += ",";
				sql += f.name;
				nr++;
			}
		}
		sql += ") VALUES (";
		nr = 0;
		for (Field f : fList) {
			if (f.isPersistent()) {
				if (nr > 0) sql += ",";
				sql += "$" + f.name + "$";
				nr++;
			}
		}
		sql += ")";
		
		sqlInsert = manager.getPool().createStatement(sql);
		
		// ------
		
		sql = "UPDATE " + tableName + " SET ";
		nr = 0;
		for (Field f : fList) {
			if (!f.isPrimary && f.isPersistent()) {
				if (nr > 0) sql += ",";
				sql += f.name + "=$" + f.name + "$";
				nr++;
			}
		}
		sql+= " WHERE ";
		nr = 0;
		for (Field f : pk) {
			sql+= (nr > 0 ? " AND " : "" ) + f.name + "=$" + f.name + "$";
			nr++;
		}
		
		sqlUpdate = manager.getPool().createStatement(sql);
		
		// ------
		
		sql = "DELETE FROM " + tableName + " WHERE ";
		nr = 0;
		for (Field f : pk) {
			sql+= (nr > 0 ? " AND " : "" ) + f.name + "=$" + f.name + "$";
			nr++;
		}
		
		sqlRemove = manager.getPool().createStatement(sql);
		
	}

	public Object getObject(DbConnection con, Object[] keys) throws Exception {

		HashMap<String, Object> attributes = new HashMap<String, Object>();
		int nr = 0;
		for (Object key : keys) {
			attributes.put(String.valueOf(nr), key);
			nr++;
		}
		DbResult ret = sqlPrimary.getStatement(con).executeQuery(attributes);
		if (!ret.next()) {
			ret.close();
			return null;
		}
		
		for (Feature f : features)
			f.getObject(con,ret);
				
		Object obj = schema.createObject(clazz, registryName, ret,manager);
		
		// fill object
		for (Field f : fList) {
			f.setToTarget(ret,obj);
		}
		
		for (Feature f : features)
			f.getObject(con,obj);
		
		return obj;
	}

	
	public void fillObject(Object obj, DbConnection con, DbResult res) throws Exception {
		
		for (Feature f : features)
			f.fillObject(obj,con,res);

		for (Field f : fList) {
			f.setToTarget(res,obj);
		}
	}
	
	public Object fillObject(DbConnection con, Object obj, Object[] keys) throws Exception {

		HashMap<String, Object> attributes = new HashMap<String, Object>();
		int nr = 0;
		for (Object key : keys) {
			attributes.put(String.valueOf(nr), key);
			nr++;
		}
		DbResult ret = sqlPrimary.getStatement(con).executeQuery(attributes);
		if (!ret.next()) {
			ret.close();
			return null;
		}
		
		for (Feature f : features)
			f.fillObject(obj, con, ret);

		// fill object
		for (Field f : fList) {
			f.setToTarget(ret,obj);
		}

		return obj;
	}
	
	/**
	 * Create the dables in the database.
	 * 
	 * @param con
	 * @throws Exception
	 */
	public void createTable(DbConnection con) throws Exception {
		
    	HashConfig cstr = new HashConfig();
    	IConfig ctable = cstr.createConfig("table");
    	ctable.setProperty("name", tableNameOrg);
    	
    	LinkedList<String> pk = new LinkedList<String>();
    	
    	for (Field f : fList) {
    		IConfig cfield = ctable.createConfig("field");
    		cfield.setProperty(Dialect.K_NAME, f.createName);
    		cfield.setProperty(Dialect.K_TYPE, f.retDbType);
    		cfield.setProperty(Dialect.K_SIZE, String.valueOf(f.size)); 
    		cfield.setProperty(Dialect.K_DEFAULT, f.defValue);
        	cfield.setProperty(Dialect.K_NOT_NULL, f.nullable ? "no" : "yes");
        	LinkedList<String> cat = new LinkedList<String>();
        	if (!f.isPersistent()) cat.add(Dialect.C_VIRTUAL);
        	if (f.isPrimary) cat.add(Dialect.C_PRIMARY_KEY);
        	if (f.ret.isEnum()) cat.add(Dialect.C_ENUMERATION);
        	cfield.setProperty(Dialect.K_CATEGORIES, MString.join(cat.iterator(), ",") );  // add primary key
    		if (f.isPrimary && f.isPersistent()) pk.add(f.createName);
    	}
    	
    	if (pk.size() > 0) {
    		String pkNames = MString.join(pk.iterator(), ",");
    		ctable.setProperty(Dialect.K_PRIMARY_KEY, pkNames);
    	}
    	
    	// create index entries
    	for (Entry<String, LinkedList<Field>> item : iIdx.entrySet()) {
	    	IConfig cindex = cstr.createConfig("index");
	    	String n = item.getKey();
	    	if (n.startsWith(DbIndex.UNIQUE)) {
	    		cindex.setString(Dialect.I_TYPE, Dialect.I_UNIQUE);
	    	}
	    	cindex.setString(Dialect.I_NAME, "idx_" + n);
	    	cindex.setString(Dialect.I_TABLE, tableNameOrg);
	    	StringBuffer fields = new StringBuffer();
	    	for (Field field : item.getValue()) {
	    		if (fields.length() > 0) fields.append(",");
	    		fields.append(field.createName);
	    	}
	    	cindex.setString(Dialect.I_FIELDS, fields.toString());
    	}
    	
    	manager.getPool().getDialect().createStructure(cstr, con, manager.getCaoMetadata());
    	
	}

	public void removeObject(DbConnection con, Object object) throws Exception {
		
		for (Feature f : features)
			f.removeObject(con,object);
	
		HashMap<String, Object> attributes = new HashMap<String, Object>();
		for (Field f : pk) {
			attributes.put(f.name, f.getFromTarget(object));
		}
		sqlRemove.getStatement(con).execute(attributes);
	}

	public String getRegistryName() {
		return registryName;
	}

	public Class<?> getClazz() {
		return clazz;
	}
	
	public String getTableName() {
		return tableNameOrg;
	}

	public Field getField(String fName) {
		return fIndex.get(fName);
	}

	public List<Field> getPrimaryKeys() {
		return pk;
	}
		
	public String toAttributes(DbPersistent pa) {
		if (pa == null) return "";
		StringBuffer out = new StringBuffer();
		
		out.append("size=").append(pa.size());
		
		if (pa.auto_id())
			out.append("&auto_id=true");

		if (!pa.nullable())
			out.append("&nullable=false");

		String type = Dialect.typeEnumToString(pa.type());
		if (!MString.isEmpty(type))
			out.append("&type=").append(type);
		
		String more = pa.more();
		if (!MString.isEmpty(more))
			out.append("&").append(more);
		return out.toString();
	}

	protected String getDbRetType(Class<?> ret) {
		String rt = Dialect.TYPE.BLOB.name();
		if (ret == int.class) rt = Dialect.TYPE.INT.name();
		else
		if (ret == long.class) rt = Dialect.TYPE.LONG.name();
		else
		if (ret == boolean.class) rt = Dialect.TYPE.BOOL.name();
		else
		if (ret == double.class) rt = Dialect.TYPE.DOUBLE.name();
		else
		if (ret == float.class) rt = Dialect.TYPE.FLOAT.name();
		else
		if (ret == String.class) rt = Dialect.TYPE.STRING.name();
		else
		if (ret == Date.class || ret == Calendar.class) rt = Dialect.TYPE.DATETIME.name();
		else
		if (ret == UUID.class) rt = Dialect.TYPE.UUID.name();
		else
		if (ret.isEnum()) rt = Dialect.TYPE.INT.name();
		
		return rt;
	}

	public String getName() {
		return name;
	}

	public List<Feature> getFeatures() {
		return features;
	}

	public IConfig getAttributes() {
		return attributes;
	}
	
}
