package de.mhus.lib.cao;

import java.util.HashMap;
import java.util.Map;

import de.mhus.lib.MActivator;
import de.mhus.lib.MEventHandler;
import de.mhus.lib.config.IConfig;

public abstract class CaoConnection implements CaoAccess, IActionProvider {

	private CaoDriver driver;
	protected HashMap<String,CaoApplication> applications = new HashMap<String, CaoApplication>();
	private Map<String, Object> conext = new HashMap<String, Object>();
	private static de.mhus.lib.logging.Log log = de.mhus.lib.logging.Log
			.getLog(CaoConnection.class);
	
	protected CaoUser user;
	
	private ChangedHandler changedHandler = new ChangedHandler();
	private IConfig config;
	protected MActivator activator;

	public CaoConnection(CaoDriver driver,IConfig config,MActivator activator) {
		this.driver = driver;
		this.config = config;
		this.activator = activator;
		driver.initializeConnection(this);
	}
	
	public CaoConnection(CaoDriver driver,IConfig config) {
		this(driver,config,driver.getActivator());
	}

	protected void setUser(CaoUser user) {
		this.user = user;
	}
	
	public boolean hasApplication(String name) {
		return applications.containsKey(name);
	}
	
	public final CaoApplication getApplication(String name) throws CaoException {
		return getApplication(null, name);
	}
	
	public CaoApplication getApplication(CaoAccess access, String name) throws CaoException {
		CaoApplication app = applications.get(name);
		if (app==null) throw new ObjectNotFoundException("Application: " + name);
		return app;
	}
	
	public final CaoDriver getDriver() {
		return driver;
	}

	public Map<String,Object> getContext() {
		return conext ;
	}

	public void fireElementUpdated(String id) {
		if (changedHandler == null) return;
		changedHandler.fireElementUpdated(id);
	}

	public void fireElementDeleted(String id) {
		if (changedHandler == null) return;
		changedHandler.fireElementDeleted(id);
	}
	
	public void fireElementCreated(String id) {
		if (changedHandler == null) return;
		changedHandler.fireElementCreated(id);
	}
	
	public void fireElementLink(String parentId, String id) {
		if (changedHandler == null) return;
		changedHandler.fireElementLink(parentId, id);
	}
	
	public void fireElementUnlink(String parentId, String id) {
		if (changedHandler == null) return;
		changedHandler.fireElementUnlink(parentId, id);
	}
	
	/**
	 * Fire a structure changed event if the internal structure of an element changed. e.g. add or remove
	 * an attribute. If all should be updated you can use null for the id.
	 * @param id The id of the changed element or null for all elements.
	 */
	public void fireElementStructurChanged(String id) {
		if (changedHandler == null) return;
		changedHandler.fireElementStructurChanged(id);
	}
	
	public void fireDisconnected() {
		if (changedHandler == null) return;
		changedHandler.fireDisconnected();
	}
	
	private class ChangedHandler extends MEventHandler<ICaoChangeListener> {

		public void fireDisconnected() {
			for (Object item : getListeners()) {
				try {
					((ICaoChangeListener)item).disconnected(CaoConnection.this);
				} catch (Throwable t) {
					log.debug(t);
				}
			}
		}

		public void fireElementUpdated(String id) {
			for (Object item : getListeners()) {
				try {
					((ICaoChangeListener)item).elementUpdated(CaoConnection.this, id);
				} catch (Throwable t) {
					log.debug(t);
				}
			}
		}
		
		public void fireElementLink(String parentId, String id) {
			for (Object item : getListenersArray()) {
				try {
					((ICaoChangeListener)item).elementLink(CaoConnection.this, parentId, id);
				} catch (Throwable t) {
					log.debug(t);
				}
			}
		}
		
		public void fireElementUnlink(String parentId, String id) {
			for (Object item : getListenersArray()) {
				try {
					((ICaoChangeListener)item).elementUnlink(CaoConnection.this, parentId, id);
				} catch (Throwable t) {
					log.debug(t);
				}
			}
		}
		
		public void fireElementCreated(String id) {
			for (Object item : getListenersArray()) {
				try {
					((ICaoChangeListener)item).elementCreated(CaoConnection.this, id);
				} catch (Throwable t) {
					log.debug(t);
				}
			}
		}

		public void fireElementDeleted(String id) {
			for (Object item : getListenersArray()) {
				try {
					((ICaoChangeListener)item).elementDeleted(CaoConnection.this, id);
				} catch (Throwable t) {
					log.debug(t);
				}
			}
		}
		
		public void fireElementStructurChanged(String id) {
			for (Object item : getListenersArray()) {
				try {
					((ICaoChangeListener)item).elementStructurChanged(CaoConnection.this, id);
				} catch (Throwable t) {
					log.debug(t);
				}
			}
		}

	}

	public CaoObserver createObserver() {
		CaoObserver observer = new CaoObserver(this);
		changedHandler.registerWeak(observer);
		return observer;
	}

	public CaoPrincipal getCurrentUser() {
		return user;
	}
	
	/**
	 * Return the config. If config is not set or null, the driver config returns.
	 * 
	 * @return
	 */
	public IConfig getConfig() {
		if (config == null)
			return getDriver().getConfig();
		return config;
	}
	
	/**
	 * Disconnect the connection will make it invalid
	 */
	public void disconnect() {
		if (!isValid()) return;
		doDisconnect();
		fireDisconnected();
		changedHandler = null;
	}
	protected abstract void doDisconnect();
	
	public abstract boolean isValid();

	public MActivator getActivator() {
		return activator;
	}
	
	public void fillWithActions(CaoElement element, CaoList targetList, CaoActionList actionList,Object...initConfig) {
	}

}
