package de.mhus.lib.cao;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import de.mhus.lib.MActivator;
import de.mhus.lib.MSingleton;
import de.mhus.lib.MString;
import de.mhus.lib.cao.util.SimpleApplicationFactory;
import de.mhus.lib.config.IConfig;
import de.mhus.lib.logging.Log;

/**
 * Sample config:
 * [cao]
 * 		[fs]
 * 
 * 			// Define a set of applications, if the config is not definet the initDefaultApplications() method will be called.
 * 			[application name="content" class="com.foo.FooClass"]
 * 				[some_app_factory_config]
 * 			[/application]
 * 			[application name="anotherone" application_class="com.foo.AppClass"] //This will use SimpleApplicationFactory, the attribute aplication_class is recommended for this factory
 * 
 * 			// If no load_application is defined then all applications will be initialized
 * 			// This is part of the create config, could be defined on another location in the config, in the config of the create connection config
 * 			[load_application name="content" alias="app1"]
 * 				[some_application_config/]
 * 			[/load_application]
 * 		[/fs]
 * [/cao]
 * 
 * @author mikehummel
 *
 */
public abstract class CaoDriver implements IActionProvider {

	public static Log log = Log.getLog(CaoDriver.class);
	
	public final static String TREE_DEFAULT = "tree_default";
	public static final String CATEGORY_ID  = "id";
	public static final String CATEGORY_ACL = "acl";
	public static final String APP_CONTENT = "content";
	public static final String LIST_DEFAULT = "list_default";

	public static final String ACTION_CREATE_FOLDER 	= "create.folder";
	public static final String ACTION_CREATE_CONTENT 	= "create.content";
	public static final String ACTION_IMPORT         	= "file.import";
	public static final String ACTION_EXPORT         	= "file.export";
	public static final String ACTION_CHECKOUT       	= "file.checkout";
	public static final String ACTION_CHECKIN 			= "file.checkin";
	public static final String ACTIVATOR_SERVICE 		= "cao.service";
	public static final String ACTION_CAP_COPY_TO 		= "cap.copy_to";
	public static final String ACTION_CAP_SET_TARGET 	= "cap.set_target";
	public static final String ACTION_DELETE 			= "file.delete";
	public static final String ACTION_CAP_MOVE_TO 		= "cap.move_to";
	public static final String ACTION_CAP_LINK_TO 		= "cap.link_to";
	
	public static final String DEFAULT_CONFIG = "cao";

	/**
	 * Implement this query in your application to allow auto completion for queries.
	 * The first attribute is the name of the query.
	 */
	public static final String AUTO_COMPLEATION = "auto_completion";
	public static final String AUTO_COMPLETION_LITERALS = "literals";
	public static final String AUTO_COMPLETION_LABELS = "labels";
	public static final String AUTO_COMPLETION_NAME = "name";

	private Map<String, CaoApplicationFactory> applications = new HashMap<String, CaoApplicationFactory>();
	private String name;
	private IConfig config;

	private MActivator activator;
	
	protected CaoActionForwarder actionForwarder = new CaoActionForwarder();
	
	/**
	 * 
	 * @param form
	 * @return
	 * @throws CaoException
	 */
	public CaoConnection createConnection(CaoForm form) throws CaoException {
		IConfig c = form.getConfig();
		if (c == null) c = getConfig();
		return createConnection(form.toUrl(true),c);
	}
	
	/**
	 * Create a connection with the given parameters and config.
	 * 
	 * @param url Must represent all login credentials to open the connection.
	 * @param config A config for the connection - this is a application depending configuration, not set by the user, It could be null also.
	 * @return
	 * @throws CaoException
	 */
	public abstract CaoConnection createConnection(String url,IConfig config) throws CaoException;
	
	public abstract CaoForm createConfiguration();
	
	/**
	 * Init the driver with default values. Init is ony one time possible.
	 * 
	 * @param name Must be set to the service name
 	 * @param config The config or null, this will use the defaut config from MSingleton /cao/<service name>
	 * @param activator A activator to load the applications or null, this will use the default activator from MSingleton 
	 * @throws AlreadyInitialized Thrown if the method is called twice
	 */
	public synchronized void initialize(String name, IConfig config, MActivator activator) throws AlreadyInitialized {
		if (this.name != null)
			throw new AlreadyInitialized(getName());
		this.name = name;
		this.config = config;
		if (activator == null) activator = MSingleton.instance().getActivator();
		this.activator = activator;
		
		config = getConfig(); // init config if needed

		log.t(getName(),"initialize",config);
		
		if (config != null) {
			for (IConfig capp : config.getConfigBundle("application") ) {
				String aName = capp.getExtracted("name");
				String aClass = capp.getExtracted("class");
				
				if (!MString.isEmpty(aName)) {
					try {
						CaoApplicationFactory aFactory = null;
						if (MString.isEmpty(aClass)) {
							aFactory = new SimpleApplicationFactory();
							((SimpleApplicationFactory)aFactory).setActivator(activator);
						} else {
								aFactory = (CaoApplicationFactory) activator.createObject(aClass);
						}
						aFactory.setConfig(capp);
						registerApplication(aName, aFactory);
					} catch (Exception e) {
						log.e(getName(),aName,aClass,"can't create object",e);
					}
				}
				
			}
		}
		
		if (applications.size() == 0) {
			initDefaultApplications();
		}
		
	}
	
	/**
	 * If the configuration do not initialize minimum one application factory. This
	 * method is called to init the the default factories.
	 */
	protected abstract void initDefaultApplications();
	
	public synchronized IConfig getConfig() {
		if (config == null) {
			IConfig c = MSingleton.instance().getConfig().getConfig(DEFAULT_CONFIG);
			if (c != null) {
				log.t(getName(),"loaded default config",config);
				config = c.getConfig(getName());
			}
		}
		return config;
	}
	
	public void registerApplication(String name, CaoApplicationFactory factory) {
		
		log.d(getName(),"register factory", name, factory);
		factory.setName(name);
		applications.put(name,factory);
	}
	
	public CaoApplicationFactory unregisterApplication(String name) {
		log.d(getName(),"unregister factory",name);
		return applications.remove(name);
	}
	
	/**
	 * This will be called by the constructor of the conection to initialize the
	 * applications.
	 * 
	 * @param con
	 * @return
	 */
	protected CaoConnection initializeConnection(CaoConnection con) {
		createApplications(con);
		return con;
	}

	/**
	 * Create all applications for the connection with help of the factories.
	 * 
	 * @param con
	 */
	private void createApplications(CaoConnection con) {
		IConfig c = con.getConfig();
		if (c!=null) {
			IConfig[] bundle = c.getConfigBundle("load_application");
			if (bundle.length == 0) {
				for (Entry<String, CaoApplicationFactory> factory : this.applications.entrySet()) {
					log.t(getName(),"Load Application",factory.getKey());
					CaoApplication app = factory.getValue().create(con);
					if (app != null) {
						log.t(getName(),"Register Application",factory.getKey());
						putApplication(con,factory.getKey(), app);
					}
				}
			} else {
				for (IConfig cload : bundle) {
					String aName = cload.getExtracted("name");
					try {
						log.t(getName(),"Load Application",aName);
						CaoApplicationFactory factory = this.applications.get(aName);
						CaoApplication app = factory.create(con, cload);
						if (app != null) {
							String alias = cload.getExtracted("alias",aName);
							log.t(getName(),"Register Application",alias,aName);
							putApplication(con,alias, app);
						}
					} catch (Throwable t) {
						log.e(getName(),"loading application",aName,t);
					}
				}
			}
		}
	}
	
	protected void putApplication(CaoConnection con, String name, CaoApplication app) {
		con.applications.put(name, app);
	}
	
	public String getName() {
		return name;
	}
	
	public MActivator getActivator() {
		return activator;
	}
	
	public void fillWithActions(CaoElement element, CaoList targetList, CaoActionList actionList,Object...initConfig) {
		
		try {
			if (element == null) {
				for (CaoElement ele : targetList.getElements()) {
					ele.getApplication().fillWithActions(ele, targetList, actionList, initConfig);
				}
			} else {
				element.getApplication().fillWithActions(element, targetList, actionList, initConfig);
			}
		} catch (Exception e) { log.t(e); }
		actionForwarder.fillWithActions(element, targetList, actionList, initConfig);
	}
	
}
