package de.mhus.lib.cao;

import java.util.Set;

import de.mhus.lib.IProperties;
import de.mhus.lib.MException;

public abstract class CaoElement extends IProperties {

	private CaoApplication app;
	protected CaoElement parent;
	
	protected CaoElement(CaoElement parent) throws CaoException {
		this(parent.getApplication());
		this.parent = parent;
	}
	
	protected CaoElement(CaoApplication app) throws CaoException {
		this.app = app;
	}

	/**
	 * return a list of children of this node. If it is not a node
	 * it will throw an exception.
	 * 
	 * @return
	 * @throws CaoException
	 */
	public final CaoList getChildren() throws CaoException {
		return getChildren(null);
	}
	
	/**
	 * return a list of children of this node. If it is not a node
	 * it will throw an exception.
	 * 
	 * @return
	 * @throws CaoException
	 */
	public abstract CaoList getChildren(CaoAccess access) throws CaoException;

	/**
	 * return true if this node can have children. If the node is only a
	 * leaf, it will return false.
	 * 
	 * @return
	 */
	public abstract boolean isNode();


	/**
	 * Returns the unique id of this object for this connection. If the
	 * object is not unique, it returns null. For joining selects the
	 * unique id could not be returned.
	 * 
	 * @return Unique ID or null
	 */
	public abstract String getId() throws CaoException;

	/**
	 * Return a display name of this object.
	 * 
	 * @return The display name
	 */
	public abstract String getName() throws CaoException;
	
	/**
	 * Returns the meta structure definition for this data.
	 * 
	 * @return The meta structure definition
	 */
	public abstract CaoMetadata getMetadata();
	
	/**
	 * Returns a parent element where this element is coming from. Usually this
	 * is possible if the element is created from a tree node. The element can 
	 * also have other parents. If no parent is known in this situation the
	 * function returns null.
	 * 
	 * @return The parent in this situation or null
	 */
	public CaoElement getParent() {
		return parent;
	}
	
	/**
	 * Returns the value of a string attribute.
	 * 
	 * @param name Name of the attribute. Throws an error if the column is not defined.
	 * 
	 * @return The value of the attribute
	 */
	public abstract String getString(String name) throws CaoException;
	
	public final CaoList getList(String name, String... attributes) throws CaoException {
		return getList(name, null, attributes);
	}
	
	public abstract CaoList getList(String name, CaoAccess access, String... attributes) throws CaoException;
	
	public abstract Object getObject(String name, String... attributes) throws CaoException;


	/**
	 * Returns true if the data is writable.
	 * 
	 * @return true if it can be written
	 */
	public abstract boolean isWritable();

	public abstract CaoWritableElement getWritableNode() throws CaoException;
	
	/**
	 * If the object is changed but not written this will recover the
	 * original state of the object. It will also load the new state from the
	 * system is something was changed from another process. After reload the
	 * object is no more dirty.
	 */
	public abstract void reload() throws CaoException;
	
	public final CaoConnection getConnection() {
		return app.getConnection();
	}
	
	public final CaoApplication getApplication() {
		return app;
	}
	
	public final CaoDriver getDriver() {
		return getConnection().getDriver();
	}
	
	public String toString() {
		try {
			return getName() + " ("+getId()+")";
		} catch (CaoException e) {
		}
		return super.toString();
	}
	
	/**
	 * Lock the object if possible. If the system do not support locking
	 * it should return false but not throw an exception. If successful
	 * locked it returns true. If the timeout is reached an exception will
	 * be thrown.
	 * 
	 * If timeout is set to 0 it will only try one time.
	 * 
	 * @param timeout Timeout in seconds. If less than 0 it will throw an exception.
	 * @throws CaoException Thrown if the timeout is reached CaoTimeoutException.
	 */
	public abstract boolean lock(int timeout) throws CaoException;

	/**
	 * Unlock the object. If unlock is not possible because the lock
	 * is not owned it will return false. If the unlock was successful
	 * it returns true.
	 * 
	 * @throws CaoException
	 */
	public abstract boolean unlock() throws CaoException;

	/**
	 * Return true if the element is locked. If locking is not supported
	 * it will always return true if you ask for your own lock.
	 * 
	 * @param owner check if the lock is for me
	 * @return
	 * @throws CaoException
	 */
	public abstract boolean isLocked(boolean owner) throws CaoException;
		
	/**
	 * Returns true if the element is valid, invalid means e.g. the corresponding
	 * element has been deleted. This method returns not false if the element data is changed. Only
	 * CaoWritable can be changed and the have a isDirty() method.
	 * 
	 * @return
	 */
	public abstract boolean isValid();

	public boolean equals(Object other) {
		if (other instanceof CaoElement) {
			try {
				return ((CaoElement)other).getConnection().equals(getConnection()) && ((CaoElement)other).getId().equals(getId());
			} catch (Throwable t) {}
		}
		return super.equals(other);
	}

	public String getString(String name, String def) {
		String value;
		try {
			value = getString(name);
		} catch (CaoException e) {
			return def;
		}
		return value==null?def:value;
	}

	@Override
	public Object getProperty(String name) {
		try {
			return getObject(name);
		} catch (CaoException e) {
		}
		return null;
	}

	@Override
	public boolean isProperty(String name) {
		try {
			getString(name);
			return true;
		} catch (CaoException e) {
		}
		return false;
	}

	@Override
	public void removeProperty(String key) throws MException {
		throw new CaoNotSupportedException();
	}

	@Override
	public void setProperty(String key, Object value) throws MException {
		throw new CaoNotSupportedException();
	}
	
	/**
	 * The feature is overwritten from IPropertie. The CaoElement is not editable. Use the writableElement to change
	 * the data.
	 */
	@Override
	public boolean isEditable() {
		return false;
	}

	/**
	 * Return the current access policy for this element.
	 * 
	 * @return
	 * @throws CaoException 
	 */
	public CaoPolicy getAccessPolicy() throws CaoException {
		return new CaoPolicy(this ,true, isWritable());
	}
	
	public Set<String> keys() {
		return getMetadata().index.keySet();
	}
	
}
