package de.mhus.lib.config;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import de.mhus.lib.IProperties;
import de.mhus.lib.MException;
import de.mhus.lib.MString;
import de.mhus.lib.logging.Log;
import de.mhus.lib.parser.CompiledString;
import de.mhus.lib.parser.DefaultScriptPart;
import de.mhus.lib.parser.StringCompiler;
import de.mhus.lib.parser.StringPart;
import de.mhus.lib.util.ArraySet;

/**
 * This interface represent a generic configuration. Configurations are
 * sets of definitions. The definitions can be bound together to a inner configuration.
 * 
 * The configuration file syntax is not declared in this interface. The
 * implementation have to interpret the stream.
 * 
 */

public abstract class IConfig extends IProperties {

	protected static Log log = Log.getLog(IConfig.class);
	
	public static final int MOVE_UP = -100;	
	public static final int MOVE_DOWN = -101;
	public static final int MOVE_FIRST = -102;
	public static final int MOVE_LAST = -103;

	private Object userObject = null;
	private ConfigStringCompiler compiler;
	private HashMap<String, CompiledString> compiledCache;
	
	/**
	 * Return the value or null if the key is not found.
	 * 
	 * @param key
	 * @return
	 */
//	@Override
//	public abstract String getProperty(String key);

	/**
	 * Return all existing keys. A property key is unique.
	 *  
	 * @return
	 */
	public abstract String[] getPropertyKeys();

	/**
	 * Returns a inner configuration by the name. Inner configurations
	 * are not unique. If more then one configurations exists it will
	 * return the first one. if no configuration exists it returns null.
	 * 
	 * @param key
	 * @return
	 */
	public abstract IConfig getConfig(String key);

	/**
	 * Return all inner configurations ignoring the name. The order
	 * is like in the configuration file. This never returns null.
	 * 
	 * @param key
	 * @return
	 */
	public abstract IConfig[] getConfigBundle();

	/**
	 * Return all inner configurations by the given name. The order
	 * is like in the configuration file. This never returns null.
	 * 
	 * @param key
	 * @return
	 */
	public abstract IConfig[] getConfigBundle(String key);

	/**
	 * Return all possible, existing inner configuration names.
	 * @return
	 */
	public abstract String[] getConfigKeys();

	/**
	 * Return a name of this config element could also be null.
	 * The name most time is the name of a sub config.
	 * @return
	 */
	public abstract String getName();

	/**
	 * Create a new config and append it at the end of the list. After creation the element is already at the end of the
	 * list.
	 * 
	 * @param key
	 * @return
	 * @throws MException if editing is not possible
	 */
	public abstract IConfig createConfig(String key) throws MException;
	
	/**
	 * Move the position of the configuration in the hole set of configurations. If the config
	 * is not part of the set it will throw an exception. The first position is 0.
	 * 
	 * @param config The config object which you want to move
	 * @param newPos Absolute new position or one of the move operators MOVE_UP, MOVE_DOWN, MOVE_FIRST, MOVE_LAST
	 * @return The new pos of the config
	 * @throws MException if editing is not possible or the position is out of range
	 */
	public abstract int moveConfig(IConfig config, int newPos) throws MException;
	
	/**
	 * Remove a config element from the set. If the config element is not part of the set it will
	 * not throw an error.
	 * 
	 * @param config The config object which should be removed.
	 * @throws MException if editing is not possible
	 */
	public abstract void removeConfig(IConfig config) throws MException;

	public void setUserObject(Object userObject) {
		this.userObject = userObject;
	}

	public Object getUserObject() {
		return userObject;
	}

	/**
	 * Returns a compiled and executed string. The string is compiled with StringCompiler and is cached. Example
	 * is "this is the value of another $anothername,default$" or with the prefix "root.": "This is a root attribute $root.name$
	 * 
	 * User the "../" prefix to go one back ...
	 * 
	 * @see StringCompiler
	 * @param key
	 * @return
	 */
	public String getExtracted(String key) {
		return getExtracted(key, null);
	}
	
	public String getExtracted(String key, String def) {
		return getExtracted(key,def,0);
	}
	
	protected String getExtracted(String key, String def,int level) {
		
		if (level > 10) return def;
		
		String value = getString(key,null);
		
		if (value == null) return def;
		if (value.indexOf('$') < 0) return value;
		
		synchronized (this) {
			if (compiler == null) {
				compiler = new ConfigStringCompiler();
				compiledCache = new HashMap<String, CompiledString>();
			}
			CompiledString cached = compiledCache.get(key);
			if (cached == null) {
				cached = compiler.compileString(value);
				compiledCache.put(key, cached);
			}
			return cached.execute(level == 0 ?null : new ConfigMap(level));
		}
	}
	
	public void setString(String name, String value) throws MException {
		synchronized (this) {
			if (compiledCache!=null)
				compiledCache.remove(name);
		}
		super.setString(name, value);
	}

	
	private class ConfigStringCompiler extends StringCompiler {
		
		protected StringPart createDefaultAttributePart(String part) {
			if (part.startsWith("root."))				
				return new RootAttributePart(part);
			if (part.startsWith(">>>"))
				return new DefaultScriptPart(part);
			return new ConfigAttributePart(part);
		}
	}
	
	private class RootAttributePart implements StringPart {
		private String name;
		private String def;
		private IConfig root;

		public RootAttributePart(String part) {
			name = MString.afterIndex(part,'.');
			root = IConfig.this;
			while (root.getParent() != null && root.getParent() != root) root = root.getParent();
			int pos = name.indexOf(',');
			if (pos > 0) {
				def = name.substring(pos+1);
				name = name.substring(0,pos);
			}
		}

		@Override
		public void execute(StringBuffer out, Map<String, Object> attributes) {
			out.append(root.getString(name,def));
		}		
		
		@Override
		public void dump(int level, StringBuffer out) {
			MString.appendRepeating(level, ' ', out);
			out.append(getClass().getCanonicalName()).append(name).append("(").append(def).append(")");
		}

	}
	
	private class ConfigAttributePart implements StringPart {

		private String name;
		private String def;
		private IConfig config;

		public ConfigAttributePart(String part) {
			name = part;
			int pos = name.indexOf(',');
			if (pos > 0) {
				def = name.substring(pos+1);
				name = name.substring(0,pos);
			}
			config = IConfig.this;
			while (name.startsWith("../")) {
				config = config.getParent();
				name = name.substring(3);
				if (config == null ) break;
			}
		}

		@Override
		public void execute(StringBuffer out, Map<String, Object> attributes) {
			int level = 0;
			if (attributes != null && attributes instanceof ConfigMap) {
				level = ((ConfigMap)attributes).getLevel();
			}
			if (config== null)
				out.append(def);
			else
				out.append(config.getExtracted(name,def,level));
		}
		
		@Override
		public void dump(int level, StringBuffer out) {
			MString.appendRepeating(level, ' ', out);
			out.append(getClass().getCanonicalName()).append(name).append("(").append(def).append(")");
		}

	}
	
	private class ConfigMap implements Map<String,Object> {

		private int level;

		private ConfigMap(int level) {
			this.level = level;
		}
		
		@Override
		public int size() {
			return 0;
		}

		public int getLevel() {
			return level;
		}

		@Override
		public boolean isEmpty() {
			return false;
		}

		@Override
		public boolean containsKey(Object key) {
			return false;
		}

		@Override
		public boolean containsValue(Object value) {
			return false;
		}

		@Override
		public Object get(Object key) {
			return null;
		}

		@Override
		public Object put(String key, Object value) {
			return null;
		}

		@Override
		public Object remove(Object key) {
			return null;
		}

		@Override
		public void putAll(Map<? extends String, ? extends Object> m) {
		}

		@Override
		public void clear() {
		}

		@Override
		public Set<String> keySet() {
			return null;
		}

		@Override
		public Collection<Object> values() {
			return null;
		}

		@Override
		public Set<java.util.Map.Entry<String, Object>> entrySet() {
			return null;
		}
		
	}

	/**
	 * return the parent config if possible.
	 * 
	 * @return
	 */
	public abstract IConfig getParent();
	
	public boolean canSave() {
		IConfig parent = getParent();
		if (parent == null) return false;
		return parent.canSave();
	}
	
	public void save() throws MException {
		IConfig parent = getParent();
		if (parent == null) {
			log.d("can't save config",this);
			return;
		}
		parent.save();
	}
	

	@Override
	public Set<String> keys() {
		return new ArraySet<String>(getPropertyKeys());
	}

	
}
