package de.mhus.lib.config;

import java.util.Collection;
import java.util.Map;
import java.util.Map.Entry;

import de.mhus.lib.util.ImmutableCollection;

public abstract class IConfigProvider {

	private Map<String, IConfig> configList;

	/**
	 * Returns a config with the given name or null if the config is not found.
	 * 
	 * @param key The config or null.
	 * @return
	 * @throws Exception 
	 */
	public IConfig getConfig(String key) throws Exception {
		synchronized(this) {
			IConfig out = configList.get(key);
			if (out == null) {
				out = createConfig(key);
				if (out != null) configList.put(key, out);
			}
			return out;
		}
	}
	
	/**
	 * Implement this function to provide configurations. The configurations will be stored
	 * in a cache object.
	 * 
	 * @param key
	 * @return
	 * @throws Exception 
	 */
	protected abstract IConfig createConfig(String key) throws Exception;

	/**
	 * Return true if the provider is also able to write changed configurations back.
	 * 
	 * @return
	 */
	public abstract boolean isWriteable();

	/**
	 * Write changed configurations back to the store.
	 */
	public void write() {
		if (!isWriteable()) return;
		synchronized(this) {
			for (Entry<String, IConfig> config : configList.entrySet()) {
				if (config.getValue() instanceof IConfigRoot) {
					IConfigRoot root = (IConfigRoot)config.getValue();
					if (root.isConfigChanged())
						writeConfiguration(config.getKey(),root);
				}
			}
		}
	}

	/**
	 * Overwrite this to allow store back of the configuration.
	 * @param string
	 * @param config
	 */
	protected abstract void writeConfiguration(String name, IConfigRoot config);
	
	/**
	 * Returns a list of loaded configurations.
	 * 
	 * @return
	 */
	public Collection<String> getLoadedKeys() {
		synchronized(this) {
			return new ImmutableCollection<String>(configList.keySet());
		}
	}
	
}
