package de.mhus.lib.form.builders;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.lang.annotation.Annotation;
import java.lang.reflect.Method;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import de.mhus.lib.MActivator;
import de.mhus.lib.MString;
import de.mhus.lib.config.IConfig;
import de.mhus.lib.config.XmlConfig;
import de.mhus.lib.form.MForm;
import de.mhus.lib.form.MFormModel;
import de.mhus.lib.form.annotations.Form;
import de.mhus.lib.form.annotations.FormAspect;
import de.mhus.lib.form.annotations.FormElement;
import de.mhus.lib.form.annotations.FormLayout;
import de.mhus.lib.form.annotations.FormSortId;
import de.mhus.lib.form.objects.FObject;

public class FormModelBuilder extends MFormModel {

	public static final String DEFAULT_SCHEMA = "";
	private SAXParser parser;
	private ElementHandler elementHandler;
	private Element root;
	private Document doc;
	private FormElementRegistry registry;
	private String schema = DEFAULT_SCHEMA;
	
//	public FormModelBuilder(MActivator activator, MForm target) throws Exception {
//		this(activator,target,"");
//	}
	
	public FormModelBuilder(MActivator activator, MForm target, String schema) throws Exception {
		super(activator, target);
		
		if (schema == null) schema = DEFAULT_SCHEMA;
		this.schema = schema;
		
		parser = SAXParserFactory.newInstance().newSAXParser();
		elementHandler = new ElementHandler();
		DocumentBuilderFactory factory = DocumentBuilderFactory
		.newInstance();
		doc = factory.newDocumentBuilder().newDocument();
		root = doc.createElement("form");
		doc.appendChild(root);
		registry = (FormElementRegistry) getActivator().getObject( MFormModel.ELEMENT_REGISTRY );

		parseClass();
		parseElements("",getTarget());
		
		parseLayout();
		
		root = null;
		elementHandler = null;
		parser = null;
		doc = null;
		registry = null;
		
		loadFromTarget();
		
	}

	private void parseLayout() throws ParserConfigurationException, SAXException, IOException {
		FormLayout layout = getTarget().getClass().getAnnotation(FormLayout.class);
		if (layout != null) {
			String xml = layout.value();
			if (!xml.startsWith("<"))
				xml = "<layout " + xml + "/>";
			
			elementHandler.setBaseElement(doc, root);
			parser.parse(new ByteArrayInputStream(xml.getBytes()), elementHandler);
			Element out = elementHandler.first;
			if (out != null) {
				setLayout( new FormLayoutBuilder(this, new XmlConfig(out)));
				return;
			}
		}
		
		setLayout( new FormLayoutSimpleBuilder(this) );

	}

	private IConfig fillElement(String prefix,Method method, String xml) throws SAXException, IOException {

		if (!xml.startsWith("<"))
			xml = "<" + xml + "/>";

		Element out = null;
		elementHandler.setBaseElement(doc, root);
		parser.parse(new ByteArrayInputStream(xml.getBytes()), elementHandler);
		out = elementHandler.first;

		if (method != null && method.getName().startsWith("set")) {
			String name = method.getName();
			name = name.substring(3, 4).toLowerCase() + name.substring(4);
			out.setAttribute("name", name);
			out.setAttribute("id", prefix + name);
		}
		return new XmlConfig(out);
	}
	
	private void parseElements(String prefix, MForm target) throws Exception {

		for (Method method : target.getClass().getMethods()) {
			
			FormAspect aspect = method.getAnnotation(FormAspect.class);
			FormElement element = method.getAnnotation(FormElement.class);
			Form form = method.getAnnotation(Form.class);
			
			if (form != null) {
				for (FormElement e : form.value()) {
					if (schema.equals(e.schema())) {
						element = e;
						break;
					}
						
				}
			}
			
			if (aspect != null) {
				
				String aspectName = aspect.value();
				MForm innerTarget = (MForm) method.invoke(target);
				if (innerTarget != null)
					parseElements(prefix + aspectName + ".", innerTarget);
				
			} else
			if (element != null) {
				FormSortId idx = method.getAnnotation(FormSortId.class);
				IConfig xmlElement = fillElement(prefix,method, toXmlString(element,method.getName()));
				FObject next = createModelObjectFromXml(xmlElement,target);
				
				if (next != null && element.sort() >= 0)
					next.setSortId(element.sort());
				else
				if (next != null && idx != null)
					next.setSortId(idx.value());
			}
		}
		
	}
	
	protected String toXmlString(FormElement element, String methodName) {
		if (!MString.isEmpty(element.value())) return element.value();
		return element.type().name().toLowerCase() + " title='" + element.title() + "' nls='" + methodName.toLowerCase() + "'";
	}

	private FObject createModelObjectFromXml(IConfig item,MForm target) throws Exception {
		
		FObject next = registry.createElement(this,item);
		if (next != null) {
			getList().add(next);
			next.initialize(this);
		}
		return next;
	}
	
	private void parseClass() throws SAXException, IOException {
		Annotation annotation = getTarget().getClass().getAnnotation(
				FormElement.class);
		if (annotation != null) {
			ClassHandler handler = new ClassHandler();
			String xml = ((FormElement) annotation).value();
			if (!xml.startsWith("<"))
				xml = "<panel " + xml + "/>";
			parser.parse(new ByteArrayInputStream(xml.getBytes()), handler);
			
			
			
			
		}
		
	}
	
	public String getSchema() {
		return schema;
	}

	public void setSchema(String schema) {
		this.schema = schema;
	}

	class ElementHandler extends DefaultHandler {

		private Document doc;
		private Element current;
		private Element first;

		public void setBaseElement(Document doc, Element rules) {
			this.doc = doc;
			current = rules;
			first = null;
		}

		@Override
		public void startElement(String uri, String localName, String name,
				Attributes attributes) throws SAXException {

			Element next = doc.createElement(name);
			for (int i = 0; i < attributes.getLength(); i++)
				next.setAttribute(attributes.getQName(i), attributes
						.getValue(i));
			if (first == null)
				first = next;
			current.appendChild(next);
			current = next;

		}

		@Override
		public void endElement(String uri, String localName, String name)
				throws SAXException {
			current = (Element) current.getParentNode();
		}

	}
	
	class ClassHandler extends DefaultHandler {

		@Override
		public void startElement(String uri, String localName, String name,
				Attributes attributes) throws SAXException {

			for (int i = 0; i < attributes.getLength(); i++) {
				String n = attributes.getQName(i);
				String v = attributes.getValue(i);
				
				if ("title".equals(n)) setTitle(v);
				else
				if ("name".equals(n)) setName(v);
				else
				if ("resource".equals(n))
					try {
						setResourceName(v);
					} catch (Exception e) {
						log().w(e);
					}
				
			}
		}

	}

}
