/*
 *  Copyright (C) 2002-2004 Mike Hummel
 *
 *  This library is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2.1 of the License, or
 *  (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package de.mhus.lib.form.objects;

import java.lang.reflect.InvocationTargetException;
import java.util.regex.Pattern;

import de.mhus.lib.MEventHandler;
import de.mhus.lib.MString;
import de.mhus.lib.config.IConfig;
import de.mhus.lib.form.FormException;
import de.mhus.lib.form.IFormDynamic;
import de.mhus.lib.form.MForm;
import de.mhus.lib.form.MFormModel;
import de.mhus.lib.form.ValidateException;
import de.mhus.lib.lang.MObject;
import de.mhus.lib.util.MNls;

/**
 * The base class for all inputs. Here are the common attributes and functions.
 * 
 * @author mike
 * 
 */
public abstract class FObject extends MObject {

	public static enum WHAT_CHANGED {
		VALUE, ENABLED, RULE_ERROR, ERROR, WRITEABLE, OTHER, TARGET
	};

	// public static enum VALIDE { VALIDE, WARN, ERROR };

	private String description;
	private String title;
	private boolean writeable = true;
	private Object userObject;
	private String name;
	private boolean changed;
	private boolean enabled = true;
	private String help = null;
	private boolean error = false;
	private MEventHandler<Listener> eventHandler = new MEventHandler<Listener>();
	private MFormModel form;
	private String currentError;
	private String ruleRegex;
	private String ruleError;

	private String id;

	private int sortId;
	private Pattern rulePattern;

	public void initialize(MFormModel pForm) {
		form = pForm;
	}

	public MFormModel getForm() {
		return form;
	}

	protected void setChanged(boolean in) {
		if (in == changed)
			return;
		changed = in;
		fireChanged(this, WHAT_CHANGED.VALUE);
	}

	public boolean isChanged() {
		return changed;
	}

	public void setCurrentError(String in) {
		if (MString.equals(in, currentError))
			return;
		currentError = in;
		fireChanged(this, WHAT_CHANGED.RULE_ERROR);
	}

	public String findCurrentError() {
		return currentError;
	}

	public void fireChanged(FObject obj, WHAT_CHANGED what) {
		Object[] list = eventHandler.getListenersArray();
		for (Object item : list)
			try {
				((Listener) item).formChangedEvent(obj, what);
			} catch (Exception e) {
				log().info(e);
			}
		if (getForm() != null)
			getForm().fireChanged(obj, what);
	}

	public Object getUserObject() {
		return userObject;
	}

	public void setUserObject(Object userObject) {
		this.userObject = userObject;
	}

	public String getDescription() {
		return description;
	}

	public String findDescription() {
		if (getNls() == null)
			return description;
		String ask = description;
		if (ask == null) ask = name + ".description=";
		return getNls().find(ask);
	}

	public void setDescription(String description) {
		this.description = description;
	}

	public String getTitle() {
		return title;
	}

	public String findTitle() {
		if (getNls() == null)
			return title;
		String ask = title;
		if (ask == null) ask = name + ".title=[" + name + "]";
		String out = getNls().find(ask);
		if (out==null) return title == null ? "[title]" : title;
		return out;
	}

	public void setTitle(String title) {
		this.title = title;
	}

	public String findHelp() {
		if (getNls() == null)
			return help;
		String ask = help;
		if (ask == null) ask = name + ".help=";
		return getNls().getString(ask,null);
	}

	public boolean isWriteable() {
		return writeable;
	}

	public void setWriteable(boolean writeable) {
		if (this.writeable == writeable)
			return;
		this.writeable = writeable;
		fireChanged(this, WHAT_CHANGED.WRITEABLE);
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	protected String getNameAsMethodSuffix() {
		return name.substring(0, 1).toUpperCase() + name.substring(1);
	}

	protected abstract void setFormValue(Object value) throws FormException;

	public abstract Object getFormValue();

	protected abstract Class<?> getFormValueClass();

	public void loadFromTarget() {
		try {
			if (getTarget() instanceof IFormDynamic) {
				setFormValue(((IFormDynamic) getTarget()).getFormValue(getName()));
				setChanged(false);
				setError(false);
				return;
			}
		} catch (Throwable e) {
			log().w(getName()," Managed", e.toString());
			return;
		}

		try {
			setFormValue(getTarget().getClass().getMethod(
					(getFormValueClass() == boolean.class ? "is" : "get")
							+ getNameAsMethodSuffix(), new Class[] {}).invoke(
					getTarget(), new Object[] {}));
			setChanged(false);
			setError(false);
		} catch (Throwable e) {
			log().w(getName(), e.toString());
		}
	}

	public void saveToTarget(boolean force) throws FormException {

		if (!force && !isChanged())
			return;

		if (!validateRule()) {
			setError(true);
			setCurrentError(ruleError);
			return;
		}
		
		if (getTarget() instanceof IFormDynamic) {
			try {
				((IFormDynamic) getTarget())
						.setFormValue(getName(), getFormValue());
				setChanged(false);
				setError(false);
				return;
			} catch (ValidateException e) {
				setError(true);
				setCurrentError(e.findMessage());
			} catch (Throwable e) {
				setError(true);
				throw new FormException(getTarget(), this, "set"
						+ getNameAsMethodSuffix(),e);
			}
		}

		try {
			getTarget().getClass().getMethod("set" + getNameAsMethodSuffix(),
					new Class[] { getFormValueClass() }).invoke(getTarget(),
					new Object[] { getFormValue() });
			setChanged(false);
			setError(false);
		} catch (InvocationTargetException e) {
			Throwable f = e.getCause();
			if (f != null && (f instanceof ValidateException)) {
				setError(true);
				setCurrentError(((ValidateException)f).findMessage());
			} else {
				setError(true);
				throw new FormException(getTarget(), this, "set"
						+ getNameAsMethodSuffix(),e);
			}
		} catch (Throwable e) {
			setError(true);
			throw new FormException(getTarget(), this, "set"
					+ getNameAsMethodSuffix(),e);
		}
	}

	public boolean validateRule() {
		if (rulePattern == null) return true;
		Object value = getFormValue();
		if (value == null) return false;
		return rulePattern.matcher(value.toString()).find();
	}

//	public void targetValidate() throws FormException {
//
//		if (!validateRule()) {
//			
//		}
//		if (getTarget() instanceof IFormDynamic) {
//			((IFormDynamic) getTarget()).validateFormAttribute(getName(),
//					getFormValueClass());
//			return;
//		}
//
//		try {
//			getTarget().getClass().getMethod("set" + getNameAsMethodSuffix(),
//					new Class[] { getFormValueClass() }).getClass();
//		} catch (Throwable e) {
//			throw new FormException(getTarget(), this,e);
//		}
//	}

	public boolean isEnabled() {
		if (!enabled) return false;
		if (form != null && !form.isLocalEnabled()) return false;
		return  true;
	}
	
	public boolean isLocalEnabled() {
		return enabled;
	}

	public void setEnabled(boolean enabled) {
		if (this.enabled == enabled)
			return;
		this.enabled = enabled;
		fireChanged(this, WHAT_CHANGED.ENABLED);
	}

	public String getHelp() {
		return help;
	}

	public void setHelp(String help) {
		this.help = help;
	}

	public boolean isError() {
		return error;
	}

	public void setError(boolean error) {
		if (this.error == error)
			return;
		this.error = error;
		fireChanged(this, WHAT_CHANGED.ERROR);
	}

	public MEventHandler<Listener> getChangeHandler() {
		return eventHandler;
	}

	public static interface Listener {

		void formChangedEvent(FObject src, WHAT_CHANGED what);

	}

	public boolean isRuleError() {
		return currentError != null;
	}

	public boolean isValide(boolean checkWarnings) {
		return !(isEnabled() && isRuleError() || isError());
	}

	public MNls getNls() {
		if (form != null)
			return form.getNls();
		return null;
	}

	public void fireStatusEvents() {
		fireChanged(this, WHAT_CHANGED.ENABLED);
		fireChanged(this, WHAT_CHANGED.ERROR);
		fireChanged(this, WHAT_CHANGED.RULE_ERROR);
	}

	public MForm getTarget() {
		return form.getTarget();
	}

	public void setId(String in) {
		id = in;
	}
	
	public String getId() {
		return id;
	}

	public void setSortId(int sort) {
		sortId = sort;
	}
	
	public int getSortId() {
		return sortId;
	}

	public void initWithConfig1(MFormModel form2, IConfig config) throws FormException {
		
		setDescription(config.getString("description",null));
		setName(config.getString("name",null));
		setId(config.getString("id",null));			
		setTitle(config.getString("title",null));
		setHelp(config.getString("help",null));
		setRulePattern(config.getString("rule_pattern",null));
		setRuleError(config.getString("rule_error",null));
		
		initWithConfig(form2, config);
	}
	public abstract void initWithConfig(MFormModel form2, IConfig config) throws FormException;

	public void setRulePattern(String pattern) {
		this.ruleRegex = pattern;
		rulePattern = null;
		if (!MString.isEmpty(ruleRegex))
			rulePattern = Pattern.compile(ruleRegex);
	}

	public String getRulePattern() {
		return ruleRegex;
	}

	public void setRuleError(String ruleError) {
		this.ruleError = ruleError;
	}

	public String getRuleError() {
		return ruleError;
	}
	
	public String toString() {
//		if (getTarget() instanceof IFormDynamic) {
//			IFormDynamic t = (IFormDynamic) getTarget();
//			try {
//				return getFormValueClass().getSimpleName() + " " + getName() + "~=" + String.valueOf( t.getFormValue(getName()) );
//			} catch (FormException e) {
//				return e.toString();
//			}
//		} else
			return getFormValueClass().getSimpleName() + " " + getName() + "=" + getFormValue();
	}

	public void fireEnabledChanged() {
		fireChanged(this, WHAT_CHANGED.ENABLED);
	}

}
