package de.mhus.lib.jmx;

import java.lang.management.ManagementFactory;
import java.lang.ref.WeakReference;
import java.util.HashMap;
import java.util.Map;

import javax.management.InstanceAlreadyExistsException;
import javax.management.MBeanServer;
import javax.management.ObjectName;

import com.sun.jdmk.comm.AuthInfo;
import com.sun.jdmk.comm.HtmlAdaptorServer;

import de.mhus.lib.MCast;
import de.mhus.lib.MHousekeeperTask;
import de.mhus.lib.MPassword;
import de.mhus.lib.MSingleton;
import de.mhus.lib.config.HashConfig;
import de.mhus.lib.config.IConfig;
import de.mhus.lib.logging.Log;

public class MRemoteManager {

	private static Log log = Log.getLog(MRemoteManager.class);
	
	private MBeanServer mbs;
	private HashMap<ObjectName,Object> registry = new HashMap<ObjectName, Object>();

	private HtmlAdaptorServer server;

	private Housekeeper housekeeper;

	public MRemoteManager() {
		housekeeper = new Housekeeper(this);
		IConfig config = MSingleton.instance().getConfig(this,new HashConfig());
		MSingleton.instance().getHousekeeper().register(housekeeper, config.getLong("housekeeper_sleep",30000), true);
	}
	
	public void register(JmxObject object) throws Exception {
		register(object,false);
	}
	
	public void register(JmxObject object,boolean weak) throws Exception {
		if (object instanceof JmxPackage) {
		  ((JmxPackage)object).open(this);
		} else {
			ObjectName on = object.getJmxObjectName();
//			if (registry.containsKey(on)) {
//				log.w("object already registered, remove first",on);
//			} else {
				register(on,object,true,weak);
				object.setJmxRegistered(true);
//			}
		}
	}

	public void unregister(JmxObject object) throws Exception {
		if (object instanceof JmxPackage) {
		  ((JmxPackage)object).close();
		} else {
			ObjectName on = object.getJmxObjectName();
			unregister(on);			
			object.setJmxRegistered(false);
		}
	}
	
	public void register(ObjectName name, Object object,boolean capsulate, boolean weak) throws Exception {
		
		Object proxy = object;
		if (capsulate)
			proxy = new MBeanProxy(object, null, this, name, weak);
		
		synchronized (this) {
			if (mbs != null)
				try {
					log.t("register",name);
					mbs.registerMBean(proxy, name);
				} catch (InstanceAlreadyExistsException e) {
					if (object instanceof JmxObject) {
						
						while (true) {
							String n = ((JmxObject)object).getJmxName();
							int pos = n.lastIndexOf('-');
							int nr = 1;
							if (pos > 0) {
								nr = MCast.toint(n.substring(pos+1),nr) + 1;
								n = n.substring(0,pos);
							}
							((JmxObject)object).setJmxName(n + "-" + nr);
							name = ((JmxObject)object).getJmxObjectName();
							if (proxy instanceof MBeanProxy) ((MBeanProxy)proxy).setName(name);
							try {
								mbs.registerMBean(proxy, name);
								log.t("renamed",name);
								break;
							} catch (InstanceAlreadyExistsException ex) {}
						}
					} else {
						log.d(name,e);
						throw e;
					}
				} catch (Exception e) {
					log.d(name,e);
					throw e;
				}
				registry.put(name, proxy);
		}
		
	}

	public void unregister(ObjectName name) {
		synchronized (this) {
			registry.remove(name);
			if (mbs != null)
				try {
					log.t("unregister",name);
					mbs.unregisterMBean(name);
				} catch (Exception e) {
					log.i(name,e);
				}
		}
		
	}
	
	public void open() {
		synchronized (this) {
			if (mbs != null) return;
			mbs = ManagementFactory.getPlatformMBeanServer();
			for (Map.Entry<ObjectName, Object> item : registry.entrySet()) {
				try {
					log.t("register",item.getKey());
					mbs.registerMBean(item.getValue(), item.getKey());
				} catch (Exception e) {
					log.i(item.getKey(),e);
				}
			}
		}
	}
	
	public void close() {
		synchronized (this) {
			if (mbs == null) return;
			for (Map.Entry<ObjectName, Object> item : registry.entrySet()) {
				try {
					log.t("unregister",item.getKey());
					mbs.unregisterMBean(item.getKey());
				} catch (Exception e) {
					log.i(item.getKey(),e);
				}
			}			
			mbs = null;
		}
	}

	public boolean isOpen() {
		return mbs != null;
	}

	public void openServer(IConfig config) {
		if (!isOpen() || server != null) return;
		server = new HtmlAdaptorServer(config.getInt("port", 1098));
		
		for (IConfig cuser : config.getConfigBundle("user")) {
			server.addUserAuthenticationInfo(new AuthInfo(cuser.getExtracted("name"), MPassword.decode(cuser.getExtracted("password")) ));
		}
		
		// TODO load user auth infos into server
		//server.setMBeanServer(mbs);
		try {
			register(new ObjectName("adaptor:proptocol=HTTP"),server,false,false);
//			mbs.registerMBean(server, new ObjectName("adaptor:proptocol=HTTP"));
			server.setMBeanServer(mbs);
		} catch (Exception e) {
			log.w(e);
			server = null;
			return;
		}
	    server.start();
	}
	
	public void closeServer() {
		if (server == null) return;
		server.stop();
		server = null;
	}

	public void check() {
		synchronized (this) {
			for (Map.Entry<ObjectName, Object> item : new HashMap<ObjectName,Object>(registry).entrySet()) {
				if (item.getValue() instanceof MBeanProxy) {
					((MBeanProxy)item.getValue()).check();
				}
			}
		}
	}
	
	private static class Housekeeper extends MHousekeeperTask {

		private WeakReference<MRemoteManager> manager;
		
		private Housekeeper(MRemoteManager manager) {
			this.manager = new WeakReference<MRemoteManager>(manager);
		}
		
		@Override
		public void doit() throws Exception {
			MRemoteManager obj = manager.get();
			if (obj == null) {
				log.t("close");
				cancel();
				return;
			}
			log.t(getClass(),"Housekeeping");
			try {
				obj.check();
			} catch (Throwable t) {
				log.t(t);
			}
		}
		
	}
}
