package de.mhus.lib.singleton;

import java.io.File;
import java.lang.reflect.Method;

import de.mhus.lib.MActivator;
import de.mhus.lib.MException;
import de.mhus.lib.MSingleton;
import de.mhus.lib.MString;
import de.mhus.lib.config.ConfigUtil;
import de.mhus.lib.config.HashConfig;
import de.mhus.lib.config.IConfig;
import de.mhus.lib.config.MConfigFactory;
import de.mhus.lib.jmx.JmxInitializer;
import de.mhus.lib.jmx.MRemoteManager;
import de.mhus.lib.lang.DynamicClassLoader;
import de.mhus.lib.logging.ConsoleFactory;
import de.mhus.lib.logging.Log;
import de.mhus.lib.logging.LogInitializer;
import de.mhus.lib.logging.MLog;
import de.mhus.lib.persistence.DefaultPersistence;
import de.mhus.lib.persistence.MPersistenceManager;

public class Standalone extends MSingleton {
	
	public static Standalone initialize(String[] args, String configFile) {

		MLog.t("Initialize");
		
		Standalone instance = new Standalone();
		
		if (args != null)
			instance.setArguments(args);
		
		instance.initConfig(configFile);
		
		setInstance(instance);
		MLog.reset();
		
		// init Activator
		instance.initActivator();
		
		// init logger
		instance.initLogger();
				
		// init jmx
		instance.initJMX();
		
		// init persistence
		instance.initPersistence();

		// more ...
		instance.initInjection();
		
		return instance;
	}

	@Override
	public synchronized IConfig getConfig() {
		if (config == null ) {
					
			Log l = new ConsoleFactory("init",System.out);
			l.setTrace(true);
			
			if (defaultConfigFile != null) {
				l.t("Load Config","default config file",defaultConfigFile);
				File f = new File(defaultConfigFile);
				try {
					if (f.exists()) {
						config = MConfigFactory.getInstance().createConfigFor(f);
					} else {
						l.t("Load Config","default config file not exists");
					}
				} catch (Exception e) {
					l.t("Load Config",e);
				}
			}
			
			getArguments();
			if (args != null) {
				String path = args.getValue("m_config_file", 0);
				if (path != null) {
					l.t("Load Config","config from args",path);
					try {
						File f = new File(path);
						if (f.exists()) {
							config = MConfigFactory.getInstance().createConfigFor(f);
						} else {
							l.t("Load Config","config from args not exists");
						}
					} catch (Exception e) {
						l.t("Load Config",e);
					}
				}
				String[] a = args.getValues("m_config_arg");
				for (String a1 : a) {
					String k = MString.beforeIndex(a1, '=');
					String v = MString.afterIndex(a1, '=');
					try {
						l.t("Load Config","Overwrite key from args",k,v);
						config.setString(k, v);
					} catch (MException e) {}
				}
			} else
			if (System.getProperty("m_config_file") != null) {
				String path = System.getProperty("m_config_file");
				if (path != null) {
					l.t("Load Config","config from env",path);
					try {
						File f = new File(path);
						if (f.exists()) {
							config = MConfigFactory.getInstance().createConfigFor(f);
						} else {
							l.t("Load Config","config from env not exists");
						}
					} catch (Exception e) {
						l.t("Load Config",e);
					}
				}				
			}
			
			if (config == null) {
				l.t("Load Config","Empty Config");
				config = new HashConfig();
			}
		}

		return config;
	}
	
	private Standalone() {
	}
	
	
	public void initConfig(String configFile) {
		if (configFile != null) {
			setDefaultConfigFile(configFile);
		}
		config = null;  // force to reload config
	}
	
	public void initActivator() {
		
		IConfig cactivator = getConfig(MActivator.class,null);
		if (cactivator == null) return;
		
		log().t("Initialize Activator");
		try {
			IConfig cloader = cactivator.getConfig("loader");
			DynamicClassLoader loader = null;
			if (cloader != null) {
				loader = new DynamicClassLoader("main");
				loader.doSetupFromConfig(cloader);
			}
			MActivator acti = new MActivator(cactivator,loader);
			activator = acti;
		} catch (Throwable t) {
			log().w(t);
		}
	}

	protected void initLogger() {
		IConfig clog = getConfig(MLog.class,null);
		if (clog == null) return;
		
		log().t("Initialize Logger");
		try {
			IConfig firstLogger = clog.getConfig("logger");
			if (firstLogger != null) {
				// create a logger
				Log log = LogInitializer.createLogFactroy(firstLogger);
				if (log != null) {
					setLogFactory(log);
				}
			}
		} catch (Throwable t) {
			log().w(t);
		}
	}
	

	protected void initPersistence() {
		IConfig conf = getConfig(MPersistenceManager.class,null);
		if (conf == null) return;
		if (!conf.getBoolean("enabled", true)) return;
		log().t("Initialize Persistence");
		
		try {
			String clazz = conf.getExtracted("class",DefaultPersistence.class.getCanonicalName());
			if (clazz != null) {
				MPersistenceManager obj = (MPersistenceManager)getActivator().createObject(clazz);
				obj.init(conf);
				if (persistence != null) persistence.close();
				persistence = obj;
			}
		} catch (Throwable t) {
			log().w(t);
		}
		
	}
	
	protected void initJMX() {
		IConfig cjmx = getConfig(MRemoteManager.class,null);
		if (cjmx == null) return;

		log().t("Initialize JMX");
		try {
			getJmxManager().open();
			JmxInitializer.initialize(cjmx);
		} catch (Throwable t) {
			log().w(t);
		}
	}

	protected void initInjection() {
		IConfig ci = getConfig().getConfig("initialize");
		if (ci == null) return;
		for (IConfig cc : ci.getConfigBundle("start")) {
			if (ConfigUtil.checkCondition(cc)) {
				String clazz = cc.getExtracted("class");
				if (clazz != null) {
					try {
						Object obj = getActivator().createObject(clazz);
						String method = cc.getExtracted("method");
						if (method != null) {
							Method me = obj.getClass().getMethod(method, new Class<?>[0]);
							me.invoke(obj, new Object[0]);
						}
					} catch (Throwable e) {
						log().e("injection",clazz,e);
					}
				}
			}
		}
	}
	
	public void reloadConfig() {
		config = null;
		fireConfigChanged();
	}
	
}
