package de.mhus.lib.sql;

import de.mhus.lib.MException;
import de.mhus.lib.parser.Parser;

/**
 * This class is a proxy for DbProvider and it is the public interface for it. It will automatic free
 * the provider in the pool if the connection is no more needed.
 * 
 * @author mikehummel
 *
 */
public interface DbConnection {

	public static final String LANGUAGE_COMMON = "common";

	/**
	 * Commit this connection changes. The system should not use the auto commit from jdbc.
	 * 
	 * @throws Exception
	 */
	public void commit() throws Exception;

	/**
	 * Returns if the user can change the database.
	 * 
	 * @return
	 * @throws Exception
	 */
	public boolean isReadOnly() throws Exception;

	/**
	 * Rollback the changes in the connection.
	 * 
	 * @throws Exception
	 */
	public void rollback() throws Exception;
	
	/**
	 * Returns a predefined statement.
	 * 
	 * @param name Name of the predefined statement.
	 * @return The statement or null if the statement is not found.
	 * @throws MException
	 */
	public DbStatement getStatement(String name) throws MException;
	
	/**
	 * Create the default language using the detected language. Default is
	 * the default language.
	 * 
	 * @param sql
	 * @return
	 * @throws MException
	 */
	public DbStatement createStatement(String sql) throws MException;
	
	/**
	 * Create and returns a new statement. Throws an exception if the language is unknown.
	 * 
	 * @param sql sql query
	 * @param language the specific language or null if the default language should be used.
	 * @return
	 * @throws MException
	 */
	public DbStatement createStatement(String sql, String language) throws MException;

	/**
	 * Create a statement from a prepared one.
	 * 
	 * @param dbPrepared
	 * @return
	 */
	DbStatement createStatement(DbPrepared dbPrepared);

	/**
	 * Returns true of the connection is no more valid.
	 * 
	 * @return
	 */
	public boolean isClosed();

	/**
	 * Returns true if the connection is in use. This is needed for the db connection pool.
	 * 
	 * @return
	 */
	public boolean isUsed();

	/**
	 * Do not use this function. It will be used by the connection pool.
	 * 
	 * @param used
	 */
	public void setUsed(boolean used);

	/**
	 * Close the connection.
	 */
	public void close();

	/**
	 * Return the unique id of this connection. Need to log information.
	 * 
	 * @return
	 */
	public long getInstanceId();

	/**
	 * Create a query compiler for this connection. Used by the DbStatement class.
	 * @return
	 * @throws MException 
	 */
	public Parser createQueryCompiler(String language) throws MException;
	
	/**
	 * Return the original instance thrue proxies - if possible.
	 * 
	 * @return
	 */
	public DbConnection instance();

	public void setUsedTrace(StackTraceElement[] createStackTrace);
	
	public StackTraceElement[] getUsedTrace();

	/**
	 * Return the default used language for this kind of connection.
	 * 
	 * @return The default supported language, could also be null as default value.
	 */
	public abstract String getDefaultLanguage();
	
	/**
	 * Return a list of supported languages. Every connection should at least
	 * support the LANGUAGE_COMMON.
	 * 
	 * @return The list of languages or an empty array but never null.
	 */
	public abstract String[] getLanguages();
	
}
