/*
 *  Copyright (C) 2002-2004 Mike Hummel
 *
 *  This library is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 2.1 of the License, or
 *  (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package de.mhus.lib.sql;

import java.lang.ref.WeakReference;
import java.util.LinkedList;
import java.util.List;

import de.mhus.lib.MActivator;
import de.mhus.lib.MException;
import de.mhus.lib.MHousekeeperTask;
import de.mhus.lib.MSingleton;
import de.mhus.lib.config.HashConfig;
import de.mhus.lib.config.IConfig;
import de.mhus.lib.jmx.JmxManaged;
import de.mhus.lib.jmx.MJmx;
import de.mhus.lib.parser.ParseException;

/**
 * The pool handles a bundle of connections. The connections should have the same
 * credentials (url, user access). Unused or closed connections will be freed after a
 * pending period.
 * 
 * @author mikehummel
 *
 */
@JmxManaged(descrition="Database pool")
public class DbPool extends MJmx {

	private List<InternalDbConnection> pool = new LinkedList<InternalDbConnection>();
	private DbProvider provider;
	private String name;
	private IConfig config;
	
	
	/**
	 * Create a new pool from central configuration.
	 * It's used the MSingleton configuration with the key of this class.
	 * 
	 * @throws Exception
	 */
	public DbPool() throws Exception {
		this(null,null);
	}
	
	/**
	 * Create a new pool from a configuration.
	 * 
	 * @param config Config element or null. null will use the central MSingleton configuration.
	 * @param activator Activator or null. null will use the central MSingleton Activator.
	 * @throws Exception
	 */
	public DbPool(IConfig config,MActivator activator) throws Exception {
		this(createProvider(config,activator));
	}
	
	protected static DbProvider createProvider(IConfig config,
			MActivator activator) throws Exception {
		
		if (config == null) config = MSingleton.instance().getConfig().getConfig(DbPool.class.getName());
		if (activator == null) activator = MSingleton.instance().getActivator();
		
		DbProvider provider = (DbProvider) activator.createObject(config.getExtracted("provider",JdbcProvider.class.getCanonicalName()));
		provider.init(config,activator);
		
		return provider;
	}

	/**
	 * Create a pool with the DbProvider.
	 * 
	 * @param provider
	 */
	public DbPool(DbProvider provider) {
		config = MSingleton.instance().getConfig(this,new HashConfig());
		setProvider(provider);
		
		Housekeeper housekeeper = new Housekeeper(this);
		MSingleton.instance().getHousekeeper().register(housekeeper, config.getLong("housekeeper_sleep",30000), true);
	}
	
	/**
	 * Set a DbProvider for this pool.
	 * 
	 * @param provider
	 */
	protected void setProvider(DbProvider provider) {
		this.provider = provider;
		name = provider.getName();
		if (name == null) name = "pool";
		name = name + MSingleton.instance().nextUniqueId();
	}
	
	/**
	 * Returns the DbProvider, it implements the database behavior and creates new connections.
	 * 
	 * @return
	 */
	public DbProvider getProvider() {
		return provider;
	}

	/**
	 * Returns the database dialect object. (Delegated to DbProvider).
	 * @return
	 */
	public Dialect getDialect() {
		return provider.getDialect();
	}
	
	/**
	 * Look into the pool for an unused DbProvider. If no one find, create one.
	 * 
	 * @param jmxName
	 * @return
	 * @throws Exception 
	 */
	public DbConnection getConnection() throws Exception {
		log().t(name,"getConnection");
		boolean foundClosed = false;
		try {
			synchronized (pool) {
				for (InternalDbConnection con : pool) {
					if (con.isClosed() || con.checkTimedOut()) {
						foundClosed = true;
					} else
					if (!con.isUsed()) {
						con.setUsed(true);
						return new DbConnectionProxy(con);
					}
				}
				InternalDbConnection con = provider.createConnection();
				con.setPool(this);
				pool.add(con);
				con.setUsed(true);
				return new DbConnectionProxy(con);
			}
		} finally {
			if (foundClosed) cleanup(false);
		}
	}

	/**
	 * Current pool size.
	 * 
	 * @return Current pool size, also pending closed connections.
	 */
	@JmxManaged(descrition="Current size of the pool")
	public int getSize() {
		synchronized (pool) {
			return pool.size();
		}
	}

	@JmxManaged(descrition="Current used connections in the pool")
	public int getUsedSize() {
		int cnt = 0;
		synchronized (pool) {
			for (DbConnection con : new LinkedList<DbConnection>(pool)) {
				if( con.isUsed()) cnt++;
			}
		}
		return cnt;
	}
	
	/**
	 * Cleanup the connection pool. Unused or closed connections will be removed.
	 * TODO new strategy to remove unused connections - not prompt, need a timeout time or minimum pool size.
	 */
	@JmxManaged(descrition="Cleanup unused connections")
	public void cleanup(boolean unusedAlso) {
		log().t(name,"cleanup");
		synchronized (pool) {
			for (InternalDbConnection con : new LinkedList<InternalDbConnection>(pool)) {
				try {
					con.checkTimedOut();
					if( unusedAlso && !con.isUsed() || con.isClosed()) {
						con.close();
						pool.remove(con);
					}
				} catch (Throwable t) {} // for secure - do not impact the thread
			}
		}		
	}

	/**
	 * Close the pool and all connections.
	 * 
	 */
	public void close() {
		if (pool == null) return;
		log().t(name,"close");
		synchronized (pool) {
			for (DbConnection con : pool) {
				con.close();
			}
			pool = null;
		}
	}
	
	@Override
	protected void finalize() throws Throwable {
		close();
		super.finalize();
	}
	
	public DbPrepared getStatement(String name) throws MException {
		String[] query = provider.getQuery(name);
		return new DbPrepared(this, query[1], query[0]);
	}
		
	/**
	 * Create a prepared statement using the default language.
	 * 
	 * @param sql
	 * @return
	 * @throws MException 
	 */
	public DbPrepared createStatement(String sql) throws MException {
		return createStatement(sql, null);
	}
	
	/**
	 * Create a new prepared statement for further use.
	 * 
	 * @param sql
	 * @param language
	 * @return
	 * @throws MException 
	 */
	public DbPrepared createStatement(String sql, String language) throws MException {
		return new DbPrepared(this, sql, language);
	}

	@JmxManaged(descrition="Unique name of the pool")
	public String getPoolId() {
		return name;
	}

	@JmxManaged(descrition="Return the usage of the connections")
	public String dumpUsage(boolean used) {
		StringBuffer out = new StringBuffer();
		synchronized (pool) {
			for (DbConnection con : pool) {
				if (!used || con.isUsed()) {
					out.append("--- ").append(con.getClass().getCanonicalName()).append("(").append(con.isUsed() ? "used" : "unused" ).append(")\n");
					StackTraceElement[] trace = con.getUsedTrace();
					if (trace != null) {
						for (StackTraceElement ste : trace)
							out.append(ste.getClassName()).append(" ").append(ste.getMethodName())
							.append("(").append(ste.getFileName()).append(":").append(ste.getLineNumber()).append(")\n");
					}
				}
			}
			pool = null;
		}
		return out.toString();
	}
	
	private static class Housekeeper extends MHousekeeperTask {

		private WeakReference<DbPool> pool;
		private String name;
		
		private Housekeeper(DbPool pool) {
			this.pool = new WeakReference<DbPool>(pool);
			this.name = pool.getPoolId();
		}
		
		@Override
		public void doit() throws Exception {
			DbPool obj = pool.get();
			if (obj == null || obj.isClosed()) {
				log.t(name,"close");
				cancel();
				return;
			}
			log.t(getClass(),name,"Housekeeping");
			obj.cleanup(false);
		}
		
	}

	public boolean isClosed() {
		return pool == null;
	}
}
