package de.mhus.lib.sql;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

import de.mhus.lib.MActivator;
import de.mhus.lib.MException;
import de.mhus.lib.MSingleton;
import de.mhus.lib.config.IConfig;

/**
 * The class holds a bundle of different database pools.
 * 
 * @author mikehummel
 *
 */
public class DbPoolBundle {

	
	private IConfig config;
	private MActivator activator;
	private Map<String, DbPool> bundle = new HashMap<String, DbPool>();
	
	/**
	 * Create a new bundle from default configuration. Load it from MSingleton with the
	 * key of this class.
	 * 
	 */
	public DbPoolBundle() {
		this(null,null);
	}
	
	/**
	 * Create a new Bundle from configuration.
	 * 
	 * @param config Config element or null. null will use the central MSingleton configuration.
	 * @param activator Activator or null. null will use the central MSingleton activator.
	 */
	public DbPoolBundle(IConfig config, MActivator activator) {

		if (config == null) config = MSingleton.instance().getConfig(DbPoolBundle.class, null);
		if (activator == null) activator = MSingleton.instance().getActivator();

		this.config = config;
		this.activator = activator;
	}
	
	public DbPool getPool(String name) throws Exception {
		
		if (bundle == null) throw new MException("Bundle already closed");

		synchronized (bundle) {
			DbPool pool = bundle.get(name);
			if (pool == null) {
				IConfig poolCon = config.getConfig(name);
				if (poolCon != null) {
					pool = new DbPool(poolCon, activator);
					bundle.put(name, pool);
				} else {
					throw new MException("pool config not found",name);
				}
			}
			return pool;
		}
	}
	
	public String[] getNames() {
		LinkedList<String> out = new LinkedList<String>();
		for (IConfig c : config.getConfigBundle()) {
			out.add(c.getName());
		}
		return out.toArray(new String[out.size()]);
	}
	
	public IConfig getConfig(String name) {
		return config.getConfig(name);
	}
	
	public IConfig getConfig() {
		return config;
	}
	
	public void reset() {
		bundle = new HashMap<String, DbPool>();
	}
	
	public void close() {
		
		if (bundle == null) return;
		
		synchronized (bundle) {
			for (DbPool pool : bundle.values())
				pool.close();
			bundle = null;
		}
	}
	
}
